// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Package test provides isolated testing for
// go.opentelemetry.io/ote/trace/internal/telemetry. It exists to ensure that
// package is tested without importing any OTLP or pdata dependencies.
package test // import "go.opentelemetry.io/otel/trace/internal/telemetry/test"
