// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Package zipkin contains an OpenTelemetry tracing exporter for Zipkin.
//
// Deprecated: The zipkin exporter is deprecated and will be removed in early 2027.
// See the blog post "[Deprecating Zipkin Exporter]".
//
// [Deprecating Zipkin Exporter]: https://opentelemetry.io/blog/2025/deprecating-zipkin-exporters/
package zipkin // import "go.opentelemetry.io/otel/exporters/zipkin"
