// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

package resource // import "go.opentelemetry.io/otel/sdk/resource"

var (
	PlatformOSDescription         = platformOSDescription
	ReadProductName               = readProductName
	ReadDisplayVersion            = readDisplayVersion
	ReadReleaseID                 = readReleaseID
	ReadCurrentMajorVersionNumber = readCurrentMajorVersionNumber
	ReadCurrentMinorVersionNumber = readCurrentMinorVersionNumber
	ReadCurrentBuildNumber        = readCurrentBuildNumber
	ReadUBR                       = readUBR
)
