// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

package otlptracehttp

// MaxResponseBodySize exposes the package-level maxResponseBodySize variable
// to allow tests to override it.
var MaxResponseBodySize = &maxResponseBodySize
